require("kara3d.vector")
require("kara3d.matrix")
require("kara3d.camera")
require("kara3d.quaternion")
require("kara3d.transform")
require("kara3d.light")
require("kara3d.shape")

-- ********** DEFINITION AND GENERIC FUNCTIONS **********

shader = {class = "shader"}
shader.__index = shader

-- Creates a new shader instance, which will call the fx functions in sequence
-- when a face is drawn
function shader.new(fx)
  local s = {fx = fx or {}}
  setmetatable(s, shader)
  return s
end

-- ********** SAMPLE SHADERS **********

-- Simple shader that applies transforms to go from model space to ass space
function shader.unlit(cull, offx, offy)
  local s = shader.new()
  s.fx = {function (f) return s:model_to_world(f, cull) end,
          function (f) return s:world_to_view(f) end,
          function (f) return s:view_to_screen(f, cull) end,
          function (f) return s:screen_to_ass(f, offx, offy) end}
  return s
end

-- Same as unlit shader, but lighting is applied
function shader.lit(cull, offx, offy)
  local s = shader.unlit(cull, offx, offy)
  s.fx[#s.fx+1] = function (f) return s:light(f) end
  return s
end

-- Same as lit shader, but a blur is applied to make edges less sharp
function shader.lit_blurred(cull, offx, offy, border, blur)
  local s = shader.lit(cull, offx, offy)
  s.fx[#s.fx+1] = function (f) return s:normal_blur(f, border, blur) end
  return s
end

-- ********** INSTANCE FUNCTIONS **********

-- Applies this shader's functions to the face f
function shader:shade(f)
  local nf = f:clone()
  for i = 1, #self.fx do
    nf = self.fx[i](nf)
    if (not nf) then return nil end
  end
  return nf
end

-- Adds or updates some fields of this shader, to use in shading functions later
function shader:update(shp, cam, mat, lgt, amb)
  self.material = mat
  self.lights = lgt
  self.ambient = amb
  self.shape = shp
  self.camera = cam
  self.mpos = shp.transform:position(true)
  self.cpos = cam.transform:position(true)
  self.model = shp.transform:matrix(true)
  self.view = cam:view()
  self.proj = cam:projection()
end

-- ********** FACE SHADING FUNCTIONS **********

-- Transforms the face f from model space to world space
-- Adds the fields world_normal and world_center to the face's data
function shader:model_to_world(f, cull)
  if (not f) then return nil end
  if (type(cull) == "nil") then cull = true end

  local c = vector.cardinal(4, 4)
  local vs = f.vertices

  local n = (self.model * (f.normal + c) - self.mpos):set_size(3):normalize()
  local c_to_v = (self.model * (vs[#vs] + c) - self.cpos):set_size(3):normalize()

  if (cull and vector.dot(c_to_v, n) > 0) then return nil end

  local center = vector.zero(3)
  local world_vertices = {}
  for i = 1, #vs do
    local v = (self.model * (vs[i] + c)):set_size(3)
    world_vertices[i] = v
    center = center + v
  end
  center = (1 / #vs) * center

  local nf = face.new(world_vertices, n, f.tags, f.data)
  nf.data.world_normal = n
  nf.data.world_center = center

  return nf
end

-- Transforms the face f from world space to view space
-- Adds the fields view_normal and view_center to the face's data
function shader:world_to_view(f)
  if (not f) then return nil end

  local c = vector.cardinal(4, 4)
  local vs = f.vertices

  local center = vector.zero(3)
  local view_vertices = {}
  for i = 1, #vs do
    local v = (self.view * (vs[i] + c)):set_size(3)
    view_vertices[i] = v
    center = center + v
  end
  center = (1 / #vs) * center

  local n = self.view * (f.normal + c + self.cpos)
  local nf = face.new(view_vertices, n, f.tags, f.data)
  nf.data.view_normal = n
  nf.data.view_center = center

  return nf
end

-- Transforms the face f from view space to screen space
-- Adds the fields screen_normal and screen_center to the face's data
function shader:view_to_screen(f, cull)
  if (not f) then return nil end
  if (type(cull) == "nil") then cull = true end

  local c = vector.cardinal(4, 4)
  local vs = f.vertices

  local in_view = false
  local center = vector.zero(3)
  local screen_vertices = {}
  for i = 1, #vs do
    local v = self.proj * (vs[i] + c)
    v = ((1 / v:w()) * v):set_size(3)
    screen_vertices[i] = v
    center = center + v
    in_view = in_view or (v:x() >= -1 and v:x() <= 1
                      and v:y() >= -1 and v:y() <= 1
                      and v:z() >= -1 and v:z() <= 1)
  end
  center = (1 / #vs) * center

  if (not in_view) then return nil end

  local n = self.proj * (f.normal + c)
  n = ((1 / n:w()) * n):set_size(3)
  local nf = face.new(screen_vertices, n, f.tags, f.data)
  nf.data.screen_normal = n
  nf.data.screen_center = center

  return nf
end

-- Transforms the face f from view screen to ass space
-- Adds the fields ass_normal, ass_center and layer to the face's data
function shader:screen_to_ass(f, xshift, yshift)
  if (not f) then return nil end

  xshift = xshift or 0
  yshift = yshift or 0

  local c = vector.cardinal(4, 4)
  local vs = f.vertices

  local xres, yres, _, _ = aegisub.video_size()
  local hxr, hyr = xres / 2, yres / 2
  local ass_matrix = matrix.new(3, 3,
    {hxr,    0,         0,
       0, -hyr,         0,
       0,    0, 999999999})

  local center = vector.zero(3)
  local ass_vertices = {}
  for i = 1, #vs do
    local v = ass_matrix * vs[i]
    ass_vertices[i] = v
    center = center + v
  end
  center = (1 / #vs) * center

  local n = ass_matrix * f.normal
  local anpos = string.format("\\an7\\pos(%s, %s)", hxr + xshift, hyr - yshift)
  local nf = face.new(ass_vertices, n, f.tags .. anpos, f.data)
  nf.data.ass_normal = n
  nf.data.ass_center = center
  nf.data.layer = math.floor(center:z())

  return nf
end

-- Calculates the face's color with a material and lights
-- Needs the fields world_normal and world_center in the face's data
-- Adds the field color to the face's data
function shader:light(f)
  if (not f) then return nil end

  local mat = self.shape.material
  local Ka = mat.ambient
  local Kd = mat.diffuse
  local Ks = mat.specular
  local shine = mat.shininess

  local N = f.data.world_normal or vector.cardinal(3, 3)
  local P = f.data.world_center or vector.zero(3)
  local V = (self.cpos - P):normalized()

  local I = Ka * self.ambient

  for i = 1, #self.lights do
    local l = self.lights[i]
    local L = (l.transform:position(true) - P):normalized()
    local R = 2 * math.max(vector.dot(L, N), 0) * N - L

    -- Diffuse component
    I = I + Kd * math.max(vector.dot(L, N), 0) * l.diffuse

    -- Specular component
    I = I + Ks * math.pow(math.max(vector.dot(R, V), 0), shine) * l.specular
  end

  local c = I:clamp(0, 255)
  local cstr = string.format("\\1c%s\\alpha%s",
    ass_color(c:x(), c:y(), c:z()),
    ass_alpha(mat.alpha)
  )

  local nf = f:clone()
  nf.tags = nf.tags .. cstr
  nf.data.color = c

  return nf
end

-- Adds a border and blur to the face, proportional to its orientation
-- Needs the fields world_normal and world_center in the face's data
function shader:normal_blur(f, border, blur)
  if (not f) then return nil end

  border = border or 0
  blur = blur or 0

  local center = f.data.world_center
  local normal = f.data.world_normal
  local color = f.data.color
  local cosf = vector.dot((self.cpos - center):normalized(), normal)
  local blurstr = string.format("\\3c%s\\bord%s\\blur%s",
    ass_color(color:x(), color:y(), color:z()),
    border * cosf, blur * cosf)

  local nf = f:clone()
  nf.tags = nf.tags .. blurstr
  return nf
end
