require("kara3d.vector")
require("kara3d.matrix")
require("kara3d.quaternion")
require("kara3d.transform")
require("kara3d.light")
require("kara3d.camera")
require("kara3d.material")
require("utils")
re = require("re")

-- ********** DEFINITION AND GENERIC FUNCTIONS **********

shape = {class = "shape"}
shape.__index = shape

-- Creates a new shape with transform t, faces f, material m and shader s
function shape.new(t, f, m, s)
  if (type(f) == "string") then f = shape.load(f) end
  local s = {transform = t, faces = f or {},
    material = m or material.new(), shader = s}
  setmetatable(s, shape)
  return s
end

-- Clones this shape
function shape:clone()
  return shape.new(self.transform, table.copy(self.faces))
end

face = {class = "face"}
face.__index = face

-- Creates a new face with vertices c, normal n, tags t and data d
function face.new(v, n, t, d)
  local f = {vertices = v or {}, normal = n or -vector.cardinal(3, 3),
    tags = t or "", data = d or {}}
  setmetatable(f, face)
  return f
end

-- Clones this face
function face:clone()
  return face.new(table.copy(self.vertices), self.normal, self.tags, table.copy(self.data))
end

-- ********** INSTANCE FUNCTIONS **********

-- Sets this face's normal to be the average of its vertices, normalized
function face:average_normal()
  local fvs = self.vertices
  self.normal = vector.zero(3)
  for i = 1, #fvs do self.normal = self.normal + fvs[i] end
  self.normal = (1 / #fvs) * self.normal
  return self
end

-- Generate a line for a face, with the indicated tags
function face:draw(sh, tags, subs, line)
  local f = sh:shade(self)
  if (not f) then return "" end

  local l = table.copy(line)
  local v = f.vertices[#f.vertices]

  l.text = string.format("{%s}{%s}{\\p1}m %s %s ",
    tags or "", f.tags or "", v:x(), v:y())

  for i = 1, #f.vertices do
    v = f.vertices[i]
    l.text = l.text .. string.format("l %s %s ", v:x(), v:y())
  end

  if (f.data.layer) then l.layer = f.data.layer end
  l.effect = "fx"
  subs.append(l)

  return ""
end

-- Generates a line for each face of this shape, with the indicated tags
function shape:draw(subs, line, tags, cam, lgt, amb)
  self.shader:update(self, cam, self.mat, lgt, amb)

  for i = 1, #self.faces do self.faces[i]:draw(self.shader, tags, subs, line) end

  return ""
end

-- ********** SHAPE CREATION FUNCTIONS **********

-- Load faces from mesh file. Only .off files supported for now, though
function shape.load(file)
  local vertices = {}
  local faces = {}
  local vertexNumber = nil
  local faceNumber = nil
  local vertexRead = 0
  local facesRead = 0
  local linenumber = 0

  for line in io.lines(file) do
    linenumber = linenumber + 1
    local blankMatch = re.match(line, "OFF|#.*")
    if blankMatch == nil then
      local matches = re.split(line, "[[:blank:]]+", true)
      if matches ~= nil then
        if vertexNumber == nil or faceNumber == nil then
            vertexNumber = tonumber(matches[1])
            faceNumber = tonumber(matches[2])
        else
          if vertexRead < vertexNumber then
              vertices[#vertices+1] = vector.new(3, {tonumber(matches[1]), tonumber(matches[2]), tonumber(matches[3])})
              vertexRead = vertexRead + 1
          else
            local faceSize = tonumber(matches[1])
            local faceVertices = {}
            for i = 1, faceSize do
                faceVertices[i] = vertices[tonumber(matches[i+1]) + 1]
            end
            local normal = vector.cross(faceVertices[2] - faceVertices[1],
                                        faceVertices[3] - faceVertices[1])
                                  :normalized()
            faces[#faces+1] = face.new(faceVertices, normal)
            facesRead = facesRead + 1
          end
        end
      end
    end
  end

  return faces
end

-- Returns the faces of a cube of length size
-- Faces are axis-aligned
-- Schematic of the vertices:
--      1________5
--     /|      /|
--  2/__|____/6 |
--  |  3|_ _|_ _|7     __ X
--  |  /    |  /     /|
-- 4|/______|/8    Z  Y
function shape.cube(size)
  local hs = size / 2
  local vs = {vector.new(3, {-hs, -hs, -hs}), vector.new(3, {-hs, -hs,  hs}),
              vector.new(3, {-hs,  hs, -hs}), vector.new(3, {-hs,  hs,  hs}),
              vector.new(3, { hs, -hs, -hs}), vector.new(3, { hs, -hs,  hs}),
              vector.new(3, { hs,  hs, -hs}), vector.new(3, { hs,  hs,  hs})}

  local left =   face.new({vs[1], vs[2], vs[4], vs[3]}, -vector.cardinal(3, 1))
  local right =  face.new({vs[5], vs[6], vs[8], vs[7]},  vector.cardinal(3, 1))
  local bottom = face.new({vs[1], vs[2], vs[6], vs[5]}, -vector.cardinal(3, 2))
  local top =    face.new({vs[3], vs[4], vs[8], vs[7]},  vector.cardinal(3, 2))
  local back =   face.new({vs[1], vs[5], vs[7], vs[3]}, -vector.cardinal(3, 3))
  local front =  face.new({vs[2], vs[6], vs[8], vs[4]},  vector.cardinal(3, 3))

  return {left, right, bottom, top, back, front}
end

-- Returns the faces of a pyramid
-- with a base with n sides, radius r, and height h
function shape.pyramid(n, r, h)
  local hh = h / 2

  local vs = {vector.new({0, hh, 0})}
  for i = 1, n do
    local angle = i/n * 2*math.pi
    vs[i+1] = vector.new({r * math.cos(angle), -hh, r * math.sin(angle)})
  end

  local fs = {face.new({}, -vector.cardinal(3, 2))}
  for i = 1, n do
    local v1, v2 = vs[i+1], vs[i < n and i+2 or 2]
    local normal = vector.cross(v2 - vs[1], v1 - vs[1]):normalize()
    fs[1].vertices[i] = vs[i+1]
    fs[i+1] = face.new({vs[1], v1, v2}, normal)
  end

  return fs
end

-- Returns the faces of a prism
-- with a base with n sides, radius r, and height h
function shape.prism(n, r, h)
  local hh = h / 2
  local up = vector.cardinal(3, 2)

  local vs = {}
  for i = 1, n do
    local angle = i/n * 2*math.pi
    local x, z = r * math.cos(angle), r * math.sin(angle)
    vs[i]   = vector.new({x, -hh, z})
    vs[n+i] = vector.new({x,  hh, z})
  end

  local fs = {face.new({}, -up), face.new({}, up)}
  for i = 1, n do
    local v1, v2 = vs[i], vs[n+i]
    local v3, v4 = vs[i < n and i+1 or 1], vs[i < n and n+i+1 or n+1]
    fs[1].vertices[i] = v1
    fs[2].vertices[i] = v2
    fs[i+2] = face.new({v1, v2, v4, v3}, (v1 + v4):normalize())
  end

  return fs
end

-- Returns the faces of a tetrahedron fitting in a sphere of radius r
function shape.tetrahedron(r)
  local y = r * math.cos(2 * math.atan(math.sqrt(2)))
  local c, s = r * math.cos(math.pi/3), r * math.sin(math.pi/3)
  local v1, v2 = vector.new({0, r, 0}), vector.new({r, y, 0})
  local v3, v4 = vector.new({-c, y, -s}), vector.new({-c, y, s})

  return {face.new({v1, v2, v3}, -v4:normalized()),
          face.new({v2, v3, v4}, -v1:normalized()),
          face.new({v3, v4, v1}, -v2:normalized()),
          face.new({v4, v1, v2}, -v3:normalized())}
end

-- Returns the faces of a cube that fits in a sphere of radius r
function shape.hexahedron(r)
  local y = r * math.cos(2 * math.atan(math.sqrt(2)))
  local c, s = r * math.cos(math.pi/3), r * math.sin(math.pi/3)
  local v1, v2 = vector.new({0, r, 0}), vector.new({r, y, 0})
  local v3, v4 = vector.new({-c, y, -s}), vector.new({-c, y, s})

  local fs = {face.new({-v1,  v2, -v3,  v4}), face.new({v1, -v2,  v3, -v4}),
              face.new({-v1,  v3, -v4,  v2}), face.new({v1, -v3,  v4, -v2}),
              face.new({-v1,  v4, -v2,  v3}), face.new({v1, -v4,  v2, -v3})}

  for i = 1, #fs do fs[i]:average_normal() end
  return fs
end

-- Returns the faces of an octahedron that fits in a sphere of radius r
function shape.octahedron(r)
  local vx = vector.new({r, 0, 0})
  local vy = vector.new({0, r, 0})
  local vz = vector.new({0, 0, r})

  local fs = {face.new({-vx, -vy, -vz}), face.new({vx, -vy, -vz}),
              face.new({-vx, -vy,  vz}), face.new({vx, -vy,  vz}),
              face.new({-vx,  vy, -vz}), face.new({vx,  vy, -vz}),
              face.new({-vx,  vy,  vz}), face.new({vx,  vy,  vz})}

  for i = 1, #fs do fs[i]:average_normal() end
  return fs
end

-- Returns the faces of a dodecahedron that fits in a sphere of radius r
function shape.dodecahedron(r)
  local a1 = math.acos(math.sqrt(5)/3)
  local a2 = 2 * math.asin(1/math.sqrt(3))

  local vs = {vector.new({0, -r, 0}), vector.new({0, r, 0})} -- poles

  local y = r * math.cos(a1)
  local t = r * math.sin(a1)
  local angle_step = 2/3*math.pi
  for i = 1, 3 do -- vertices adjacent to the poles
    local angle = i * angle_step
    local x, z = t * math.cos(angle), t * math.sin(angle)
    vs[i+2] = vector.new({ x,  y,  z})
    vs[i+5] = vector.new({-x, -y, -z})
  end

  y = r * math.cos(a2)
  t = r * math.sin(a2)
  angle_step = math.pi/3
  for i = 1, 6 do -- vertices 2-adjacent to the poles
    local angle = i * angle_step
    local x1, z1 = t * math.cos(angle - a1/2), t * math.sin(angle - a1/2)
    local x2, z2 = t * math.cos(angle + a1/2), t * math.sin(angle + a1/2)
    local ys = i % 2 == 0 and -y or y
    vs[2*i+7] = vector.new({x1, ys, z1})
    vs[2*i+8] = vector.new({x2, ys, z2})
  end

  local fs = {}
  for i = 1, 3 do
    local j = (i % 3) + 1
    local k = (j % 3) + 1
    fs[i] = face.new({vs[1], vs[i+5], vs[4*k+7], vs[4*k+8], vs[j+5]})
    fs[i+3] = face.new({vs[2], vs[i+2], vs[4*j+5], vs[4*j+6], vs[j+2]})
    fs[i+6] = face.new({vs[i+2], vs[4*j+5], vs[4*i+8], vs[4*i+7], vs[4*i+6]})
    fs[i+9] = face.new({vs[i+5], vs[4*k+7], vs[4*k+6], vs[4*k+5], vs[4*j+8]})
  end

  for i = 1, #fs do fs[i]:average_normal() end
  return fs
end

-- Returns the faces of an icosahedron fitting in a sphere of radius r
function shape.icosahedron(r)
  local up = vector.cardinal(3, 2)
  local lat = math.atan2(1, 2)
  local t = r * math.cos(lat)
  local y = r * math.sin(lat)

  local vs = {-r * up, r * up}
  for i = 1, 5 do
    local lon1 = (i-0.5)/5 * 2*math.pi
    local lon2 = i/5 * 2*math.pi
    vs[i+2] = vector.new({t * math.cos(lon1), -y, t * math.sin(lon1)})
    vs[i+7] = vector.new({t * math.cos(lon2), y, t * math.sin(lon2)})
  end

  local fs = {}
  for i = 1, 5 do
    local vb1, vb2 = vs[i+2], vs[i < 5 and i+3 or 3]
    local vt1, vt2 = vs[i+7], vs[i < 5 and i+8 or 8]
    fs[i] = face.new({vs[1], vb1, vb2})
    fs[i+5] = face.new({vs[2], vt1, vt2})
    fs[i+10] = face.new({vb1, vb2, vt1})
    fs[i+15] = face.new({vt1, vt2, vb2})
  end

  for i = 1, #fs do fs[i]:average_normal() end
  return fs
end

-- Returns the faces of a sphere approximation with radius r
-- The generation works by subdividing the faces of the model d times,
-- starting from an icosahedron
function shape.icosphere(r, d)
  r = r or 1
  d = d or 0

  local subdivide = function(f)
    local v1, v2, v3 = f.vertices[1], f.vertices[2], f.vertices[3]
    local v12 = r * (v1 + v2):normalize()
    local v23 = r * (v2 + v3):normalize()
    local v31 = r * (v3 + v1):normalize()
    return face.new({v12, v23, v31}), face.new({v1, v12, v31}),
           face.new({ v2, v12, v23}), face.new({v3, v31, v23})
  end

  local fs = shape.icosahedron(r)
  for i = 1, d do
    local n = #fs
    for j = 1, n do
      fs[j], fs[n+j], fs[2*n+j], fs[3*n+j] = subdivide(fs[j])
    end
  end

  for i = 1, #fs do fs[i]:average_normal() end
  return fs
end
