require("kara3d.vector")
require("kara3d.matrix")
require("kara3d.quaternion")

-- ********** DEFINITION AND GENERIC FUNCTIONS **********

transform = {class = "transform"}
transform.__index = transform

-- Creates a new transform with position p, rotation r, scale s and parent par
-- p and s should be of size 3, r is a quaternion, and par may be nil
function transform.new(p, r, s, par)
  p = p and vector.new(p) or vector.zero(3)
  r = r and quaternion.new(r) or quaternion.identity()
  s = s and vector.new(s) or vector.one(3)
  local t = {pos = p, rot = r, scl = s, parent = par}
  setmetatable(t, transform)
  return t
end

-- Clones this transform
function transform:clone()
  return transform.new(self.pos:clone(), self.rot:clone(), self.scl:clone(), self.parent)
end

-- Returns a string that represents this transform
function transform:tostring()
  return "(pos = (" .. self.pos:tostring() .. "), "
    .. "rot = (" .. self.rot:tostring() .. "), "
    .. "scl = (" .. self.scl:tostring() .. "), "
    .. (self.parent and "has parent" or "no parent") .. ")"
end

-- ********** INSTANCE FUNCTIONS **********

function transform:right()
  return (self:rotation(true):to_matrix() * vector.new({1, 0, 0, 1})):set_size(3):normalize()
end

function transform:up()
  return (self:rotation(true):to_matrix() * vector.new({0, 1, 0, 1})):set_size(3):normalize()
end

function transform:forward()
  return (self:rotation(true):to_matrix() * vector.new({0, 0, 1, 1})):set_size(3):normalize()
end

-- Rotates this transform by angle around axis (local)
function transform:rotate(axis, angle)
  self.rot = quaternion.mul_quaternion(quaternion.axis_angle(axis, angle), self.rot)
  return self
end

-- Returns the local or global position of this transform
function transform:position(global)
  local p = self.pos:clone()
  if (p.size > 4) then p:set_size(4) end
  if (p.size < 4) then p:set_size(4) p:set(4, 1) end
  if (global and self.parent) then
    p = matrix.mul_vector(self.parent:matrix(true), p)
  end
  return p:set_size(3)
end

-- Returns the local or global rotation of this transform
function transform:rotation(global)
  local q = self.rot:normalized()
  if (global and self.parent) then
    q = quaternion.mul_quaternion(self.parent:rotation(true), q):normalize()
  end
  return q
end

-- Returns the local or global scale of this transform
function transform:scale(global)
  local s = self.scl:clone()
  if (global and self.parent) then
    local m = self:matrix(true)
    s:set(1, m:column(1):set_size(3):length())
    s:set(2, m:column(2):set_size(3):length())
    s:set(3, m:column(3):set_size(3):length())
  end
  return s
end

-- Returns the transformation matrix corresponding to this transform
function transform:matrix(global)
  local translation = matrix.new(4, 4,
    {1, 0, 0, self.pos:get(1),
     0, 1, 0, self.pos:get(2),
     0, 0, 1, self.pos:get(3),
     0, 0, 0,               1})
  local rotation = self.rot:normalized():to_matrix()
  local scale = matrix.new(4, 4,
    {self.scl:get(1),               0,               0, 0,
                   0, self.scl:get(2),               0, 0,
                   0,               0, self.scl:get(3), 0,
                   0,               0,               0, 1})
  local m = translation * rotation * scale

  if (global and self.parent) then
    m = self.parent:matrix(true) * m
  end

  return m
end

-- Returns the inverse of the transformation matrix of this transform
function transform:inv_matrix(global)
  local translation = matrix.new(4, 4,
    {1, 0, 0, -self.pos:get(1),
     0, 1, 0, -self.pos:get(2),
     0, 0, 1, -self.pos:get(3),
     0, 0, 0,               1})
  local rotation = self.rot:inv():normalized():to_matrix()
  local scale = matrix.new(4, 4,
    {1/self.scl:get(1),                 0,                 0, 0,
                     0, 1/self.scl:get(2),                 0, 0,
                     0,                 0, 1/self.scl:get(3), 0,
                     0,                 0,                 0, 1})
  local m = scale * rotation * translation

  if (global and self.parent) then
    m = m * self.parent:inv_matrix(true)
  end

  return m
end
